import random
import argparse
import sys, string, glob, re, os
import numpy as np
import scipy as sp
import time,copy

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import TensorDataset,DataLoader
from torch.nn import functional as F

from model_meta_cate import Cate,NeuralNet

def traineval_epoch_first(dataloader,model,optimizer,args,is_train):
    if is_train == 1:
        model.train()
    else:
        model.eval()

    total_loss = 0.0
    total_cnt = 0
    for batch_idx,(x,y,a) in enumerate(dataloader):
        loss = model.forward(x,y,a)
        if is_train == 1:
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        total_loss += loss.detach().item()
        total_cnt += 1
    return total_loss/total_cnt

def traineval_epoch_second(dataloader,model,optimizer,args,is_train):
    if is_train == 1:
        model.train()
    else:
        model.eval()

    total_loss = 0.0
    total_cnt = 0
    for batch_idx,(x,ytilde) in enumerate(dataloader):
        loss = model(x,ytilde)
        if is_train == 1:
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        total_loss += loss.detach().item()
        total_cnt += 1
    return total_loss/total_cnt

def train_large(x,y,a,tau,args):
    start = time.time()
    
    n_task,n_sample,n_att = x.shape
    if args.n_sample > 0 and args.n_sample < n_sample:
        n_sample = args.n_sample
    
    r_train = 0.7
    r_valid = 0.1
    n_train = int(n_sample*r_train)    
    n_valid = int(n_sample*r_valid)
    n_test = n_sample-n_train-n_valid
        
    tauhat = torch.zeros(n_task,n_sample,1)
    total_tauloss = 0.0
    for i in range(n_task):
        x_valid = x[i,n_test:n_test+n_valid,:]
        x_train = x[i,n_test+n_valid:n_test+n_valid+n_train,:]
        y_valid = y[i,n_test:n_test+n_valid,:]
        y_train = y[i,n_test+n_valid:n_test+n_valid+n_train,:]
        a_valid = a[i,n_test:n_test+n_valid,:]
        a_train = a[i,n_test+n_valid:n_test+n_valid+n_train,:]
        
        traindataset = TensorDataset(x_train,y_train,a_train)
        validdataset = TensorDataset(x_valid,y_valid,a_valid)
        trainloader = DataLoader(traindataset,batch_size=args.batch_size,shuffle=True)
        validloader = DataLoader(validdataset,batch_size=args.batch_size,shuffle=False)
        
        model_first = Cate(n_att=n_att,n_hiddens=args.n_hiddens,dropout=args.dropout,method=args.method,second_method=args.second_method)
        model_first.to(args.device)
        optimizer_first = torch.optim.Adam(model_first.parameters(),lr=args.lr,weight_decay=args.weight_decay)
        valid_losss = []
        bestmodel_first = copy.deepcopy(model_first)
        for epoch_i in range(args.epoch):
            train_loss = traineval_epoch_first(trainloader,model_first,optimizer_first,args,is_train=1)
            valid_loss = traineval_epoch_first(validloader,model_first,optimizer_first,args,is_train=0)
            if args.v > 0 and epoch_i % args.v == 0:
                print("first: %d %d %f %f"%(i,epoch_i,train_loss,valid_loss))
            valid_losss += [valid_loss]
            if valid_loss <= min(valid_losss):
                bestmodel_first = copy.deepcopy(model_first)


        if args.second_method < 3:
            ytilde = bestmodel_first.get_ytilde(x[i,:,:],y[i,:,:],a[i,:,:],second_method=args.second_method).detach()
            traindataset = TensorDataset(x_train,ytilde[n_test+n_valid:n_test+n_valid+n_train,:])
            validdataset = TensorDataset(x_valid,ytilde[n_test:n_test+n_valid,:])
            trainloader = DataLoader(traindataset,batch_size=args.batch_size,shuffle=True)
            validloader = DataLoader(validdataset,batch_size=args.batch_size,shuffle=False)                

            model_second = NeuralNet(n_att=n_att,n_out=1,n_hiddens=args.n_hiddens,dropout=args.dropout,binary=0)
            model_second.to(args.device)
            optimizer_second = torch.optim.Adam(model_second.parameters(),lr=args.lr,weight_decay=args.weight_decay)
            valid_losss = []
            bestmodel_second = copy.deepcopy(model_second)
            for epoch_i in range(args.epoch):
                train_loss = traineval_epoch_second(trainloader,model_second,optimizer_second,args,is_train=1)
                valid_loss = traineval_epoch_second(validloader,model_second,optimizer_second,args,is_train=0)
                if args.v > 0 and epoch_i % args.v == 0:                
                    print("second: %d %d %f %f"%(i,epoch_i,train_loss,valid_loss))
                valid_losss += [valid_loss]
                if valid_loss <= min(valid_losss):
                    bestmodel_second = copy.deepcopy(model_second)

            tauhat[i,:,:] = bestmodel_second.f(x[i,:n_sample,:])
        else:
            tauhat[i,:,:] = bestmodel_first.get_ytilde(x[i,:n_sample,:],y[i,:n_sample,:],a[i,:n_sample,:],second_method=3)
                    
        tauloss = torch.mean((tauhat[i,:n_test,:]-tau[i,:n_test,:])**2)
        total_tauloss += tauloss
        print("task: %d tauloss: %f"%(i,tauloss))

    train_time = (time.time()-start)
    print("mean-tauloss: %f train-time: %f"%(total_tauloss/n_task,train_time))
    return tauhat

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--fn',type=str,default="",help='data file, e.g., data/csv/PenDigits_withoutdupl_norm_v01_t100/I10/fns0.npz')
    parser.add_argument('--epoch','-e',type=int,default=500,help='epochs')
    parser.add_argument('--v',type=int,default=0,help='vervoise')    
    parser.add_argument('--n_hiddens',type=str,default='256,256',help='#hidden units')
    parser.add_argument('--weight_decay',type=float,default=1e-5)
    parser.add_argument('--dropout', type=float, default=0.1)
    parser.add_argument('--method',type=int,default=0,help='method')
    parser.add_argument('--second_method',type=int,default=0,help='0:RA 1:PA 2:DR 3:plugin')
    parser.add_argument('--seed',type=int,default=0,help='random seed')
    parser.add_argument('--batch_size', type=int, default=256)
    parser.add_argument('--no_cuda', action='store_true')
    parser.add_argument('--lr', type=float, default=1e-3)
    parser.add_argument('--initmodel', default=None)
    parser.add_argument('--out', default=None)
    parser.add_argument('--n_sample',type=int,default=0,help='number of samples per task')
                        
    args = parser.parse_args()    
    args.cuda = not args.no_cuda and torch.cuda.is_available()
    device = torch.device("cuda" if not args.no_cuda and torch.cuda.is_available() else "cpu")
    args.device = device
    np.random.seed(args.seed)
    random.seed(args.seed)
    torch.manual_seed(args.seed)

    print(args)

    dat = np.load(args.fn)
    x = dat['x']
    y = dat['y']
    a = dat['a']
    mu = dat['mu']    
    tau = dat['tau']
    n_task,n_sample,n_att = x.shape
    x = torch.FloatTensor(x).to(args.device)
    y = torch.FloatTensor(y).to(args.device)
    a = torch.FloatTensor(a).to(args.device)
    tau = torch.FloatTensor(tau).to(args.device)
    args.n_hiddens = list(map(int,args.n_hiddens.split(',')))        
    tauhat = train_large(x,y,a,tau,args).detach().numpy()

    if args.out is not None:
        ofn = re.sub("model","data",args.out)
        ofn = "%s.npz"%(ofn)
    print(ofn)
    np.savez(ofn,x=x,y=y,a=a,mu=mu,tau=tau,tauhat=tauhat)    
